"""Wrapper for KBW C API."""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from ctypes import POINTER, c_void_p, c_size_t, c_bool, c_int32, c_uint32, c_uint8
from functools import reduce
from operator import iconcat
from typing import Literal
from os import environ
from os.path import dirname
from .wrapper import load_lib, os_lib_name

api_argtypes = {
    "kbw_set_log_level": ([c_uint32], []),
    "kbw_build_info": ([], [POINTER(c_uint8), c_size_t]),
    "kbw_make_configuration": (
        [
            c_size_t,  # num_qubits
            c_int32,  # simulator
            c_bool,  # use_live
            POINTER(c_size_t),  # coupling_graph
            c_size_t,  # coupling_graph_size
            c_bool,  # gradient
            c_size_t,  # sample_base
            POINTER(c_size_t),  # classical_shadows
        ],
        [c_void_p],
    ),
}


def kbw_path():
    """Get KBW path"""

    if "KBW_PATH" in environ:
        path = environ["KBW_PATH"]
    else:
        path = f'{dirname(__file__)}/libs/{os_lib_name("kbw")}'

    return path


API = load_lib("KBW", kbw_path(), api_argtypes, "kbw_error_message")


def set_log(level: int):
    """Set KBW log level"""

    API["kbw_set_log_level"](level)


_SIMULATOR = {
    "dense": 0,
    "dense v1": 0,
    "sparse": 1,
    "dense v2": 2,
}


_CLASSICAL_SHADOWS = {"bias": (1, 1, 1), "samples": 1_000, "shots": 2048}


def get_simulator(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    num_qubits: int,
    execution: Literal["live", "batch"] = "live",
    simulator: Literal["sparse", "dense", "dense v2"] = "sparse",
    coupling_graph: list[tuple[int, int]] | None = None,
    gradient: bool = False,
    classical_shadows=None,
    direct_sample=0,
):
    """Create a configuration"""
    if gradient:
        execution = "batch"

    coupling_graph_size = len(coupling_graph) if coupling_graph else 0

    if coupling_graph:
        qubit_in_graph = [q for edge in coupling_graph for q in edge]
        if any(q not in qubit_in_graph for q in range(num_qubits)) or any(
            q not in list(range(num_qubits)) for q in qubit_in_graph
        ):
            raise ValueError("Unreachable qubit in the coupling graph.")
        coupling_graph = reduce(iconcat, coupling_graph, [])
        coupling_graph = (c_size_t * len(coupling_graph))(*coupling_graph)

    if classical_shadows is not None:
        cs = {**_CLASSICAL_SHADOWS, **classical_shadows}
        classical_shadows = (c_size_t * 5)(
            cs["bias"][0], cs["bias"][1], cs["bias"][2], cs["samples"], cs["shots"]
        )

    return API["kbw_make_configuration"](
        num_qubits,  # num_qubits
        _SIMULATOR[simulator.lower()],  # simulator
        execution == "live",  # use_live
        coupling_graph,  # coupling_graph
        coupling_graph_size,  # coupling_graph_size
        gradient,  # gradient
        direct_sample,  # sample_base
        classical_shadows,  # classical_shadows
    )



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from math import pi, sqrt, cos, sin
from cmath import isclose, exp
import ket
import pytest


def validate(matrix):
    gate = ket.qulib.dump_matrix(ket.qulib.gates.unitary(matrix, True))

    return all(
        isclose(gate[i][j], matrix[i][j], abs_tol=1e-10)
        for i in range(len(matrix))
        for j in range(len(matrix[0]))
    )


X = [[0, 1], [1, 0]]
Y = [[0, -1j], [1j, 0]]
Z = [[1, 0], [0, -1]]
H = [[1 / sqrt(2), 1 / sqrt(2)], [1 / sqrt(2), -1 / sqrt(2)]]
T = [[1, 0], [0, exp(1j * pi / 4)]]
S = [[1, 0], [0, 1j]]
RX = lambda theta: [
    [cos(theta / 2), -1j * sin(theta / 2)],
    [-1j * sin(theta / 2), cos(theta / 2)],
]
RY = lambda theta: [
    [cos(theta / 2), -sin(theta / 2)],
    [sin(theta / 2), cos(theta / 2)],
]
RZ = lambda theta: [[exp(-1j * theta / 2), 0], [0, exp(1j * theta / 2)]]


def test_unitary():
    for gate in [
        X,
        Y,
        Z,
        H,
        T,
        S,
        *[RX(2 * pi / i) for i in range(1, 11)],
        *[RY(2 * pi / i) for i in range(1, 11)],
        *[RZ(2 * pi / i) for i in range(1, 11)],
    ]:
        assert validate(gate)


def test_not_unitary():
    with pytest.raises(ValueError):
        sqrt_2 = 1 / sqrt(2)
        ket.qulib.gates.unitary([[sqrt_2, sqrt_2], [sqrt_2, sqrt_2]])


def test_not_2x2_matrix():
    with pytest.raises(ValueError):
        ket.qulib.gates.unitary([[1, 2], [4, 5, 6]])


def make_ctrl_gate(matrix):
    return [
        [1, 0, 0, 0],
        [0, 1, 0, 0],
        [0, 0, matrix[0][0], matrix[0][1]],
        [0, 0, matrix[1][0], matrix[1][1]],
    ]


def ctrl_validate(matrix):
    gate = lambda q: ket.ctrl(q[0], ket.qulib.gates.unitary(matrix))(q[1])
    gate = ket.qulib.dump_matrix(gate, num_qubits=2)

    matrix = make_ctrl_gate(matrix)

    return all(
        isclose(gate[i][j], matrix[i][j], abs_tol=1e-10)
        for i in range(len(matrix))
        for j in range(len(matrix[0]))
    )


def test_ctrl_unitary():
    for gate in [X, Y, Z, H, T, S]:
        assert ctrl_validate(gate)


if __name__ == "__main__":
    test_unitary()
    print("OK")



"""Hamiltonian library."""

# SPDX-FileCopyrightText: 2025 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from ..gates import X, Y, Z, obs
from ..base import Quant
from ..expv import Hamiltonian, commutator
from ..operations import exp_value

__all__ = [
    "maxcut",
    "x_mixer",
    "xy_mixer",
    "qubo",
    "falqon_a",
    "falqon_b",
    "falqon_a",
    "falqon_get_beta_fo",
    "falqon_get_beta_so",
]


def maxcut(edges: list[tuple[int, int]], qubits: Quant) -> Hamiltonian:
    r"""Max-Cut Hamiltonian.

    .. math::
        \sum_{a, b\,\in\, \mathcal{E}}\frac{1}{2}(1-Z_a Z_b)

    Args:
        edges: List of edges in the graph.
        qubits: Qubits representing the graph nodes.
    """
    with obs():
        return 1 / 2 * sum(1 - Z(a) * Z(b) for a, b in map(qubits.at, edges))


def x_mixer(qubits: Quant) -> Hamiltonian:
    r"""X-Mixer Hamiltonian.

    .. math::
        \sum_j X_j

    Args:
        qubits: Qubits to apply the mixer.
    """
    with obs():
        return sum(X(q) for q in qubits)


def xy_mixer(qubits, edges: list[tuple[int, int]] | None = None) -> Hamiltonian:
    r"""XY-Mixer Hamiltonian.

    .. math::
        \frac{1}{2}\sum_{a, b\,\in\, \mathcal{E}} X_a X_b + Y_a Y_b


    If edges is None, a ring topology is assumed.

    Args:
        qubits: Qubits to apply the mixer.
        edges: List of edges in the graph.
    """
    if edges is None:
        n = len(qubits)
        edges = [(i, (i + 1) % n) for i in range(n)]

    with obs():
        return (
            sum(X(a) * X(b) for a, b in map(qubits.at, edges))
            + sum(Y(a) * Y(b) for a, b in map(qubits.at, edges))
        ) / 2


def qubo(model, qubits: Quant) -> Hamiltonian:
    """Convert a QUBO model to a Hamiltonian.

    Converts a QUBO model from the pyQUBO library to a Hamiltonian observable.

    Args:
        model: QUBO model.
        qubits: Qubits representing the variables in the model.
    """
    linear, quadratic, offset = model.to_ising(index_label=True)

    with obs():
        return (
            offset
            + sum(c * -Z(qubits[i]) for i, c in linear.items())
            + sum(c * Z(qubits[i]) * Z(qubits[j]) for (i, j), c in quadratic.items())
        )


def falqon_a(hp: Hamiltonian, hd: Hamiltonian) -> Hamiltonian:
    """FALQON A operator.

    .. math::
        A = i[H_d, H_p]


    See https://arxiv.org/abs/2103.08619.

    Args:
        hp: Problem Hamiltonian.
        hd: Driver Hamiltonian.
    """
    return 1j * commutator(hd, hp)


def falqon_b(hp: Hamiltonian, hd: Hamiltonian) -> Hamiltonian:
    """FALQON B operator.

    .. math::
        B = [ [H_d, H_p], H_d ]


    See https://arxiv.org/abs/2407.17810.

    Args:
        hp: Problem Hamiltonian.
        hd: Driver Hamiltonian.
    """
    return commutator(commutator(hd, hp), hd)


def falqon_c(hp: Hamiltonian, hd: Hamiltonian) -> Hamiltonian:
    """FALQON C operator.

    .. math::
        C = [ [H_d, H_p], H_p ]

    See https://arxiv.org/abs/2407.17810.

    Args:
        hp: Problem Hamiltonian.
        hd: Driver Hamiltonian.
    """
    return commutator(commutator(hd, hp), hp)


def falqon_get_beta_fo(hp: Hamiltonian, hd: Hamiltonian) -> float:
    """Get FALQON first-order beta parameter.

    See https://arxiv.org/abs/2103.08619.

    This function computes the expectation value, triggering the circuit execution.

    Args:
        hp: Problem Hamiltonian.
        hd: Driver Hamiltonian.
    """
    return -exp_value(falqon_a(hp, hd)).get()


def falqon_get_beta_so(delta_t, hp: Hamiltonian, hd: Hamiltonian) -> float:
    """Get FALQON second-order beta parameter.

    See https://arxiv.org/abs/2407.17810.

    This function computes the expectation values, triggering the circuit execution.

    Args:
        delta_t: Time step.
        hp: Problem Hamiltonian.
        hd: Driver Hamiltonian.
    """
    a = exp_value(falqon_a(hp, hd))
    b = exp_value(falqon_b(hp, hd))
    c = exp_value(falqon_c(hp, hd))

    a = a.get()
    b = b.get()
    c = c.get()

    so = -(a + delta_t * c) / (2 * delta_t * b)

    return -a if abs(so) > abs(a) else so



"""Quantum gate definitions.

All quantum gates take one or more :class:`~ket.base.Quant` as input and return them at the end.
This allows for concatenating quantum operations.

Example:

    .. code-block:: python

        from ket import *

        p = Process()
        a, b = p.alloc(2)

        S(X(a))  # Apply a Pauli X followed by an S gate on `a`.

        CNOT(H(a), b)  # Apply a Hadamard on `a` followed by a CNOT gate on `a` and `b`.

For gates that take classical parameters, such as rotation gates, if non-qubits are passed,
it will return a new gate with the classical parameter set.

Example:

    .. code-block:: python

        from math import pi
        from ket import *

        s_gate = P(pi/2)
        t_gate = P(pi/4)

        p = Process()
        q = p.alloc()

        s_gate(q)
        t_gate(q)
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from contextlib import contextmanager
from math import pi
from functools import reduce
from operator import add
from typing import Any, Callable
import functools
import warnings
import contextvars

from .clib.libket import (
    HADAMARD,
    PAULI_X,
    PAULI_Y,
    PAULI_Z,
    ROTATION_X,
    ROTATION_Y,
    ROTATION_Z,
    PHASE_SHIFT,
)

from .base import Process, Quant, Parameter
from .operations import _search_process, control, ctrl, cat, kron, around
from .expv import Pauli, Hamiltonian

__all__ = [
    "I",
    "X",
    "Y",
    "Z",
    "H",
    "RX",
    "RY",
    "RZ",
    "PHASE",
    "P",
    "S",
    "T",
    "SD",
    "TD",
    "U3",
    "CNOT",
    "CZ",
    "SWAP",
    "RXX",
    "RZZ",
    "RYY",
    "SX",
    "global_phase",
    "RBS",
    "ham",
    "obs",
    "QFT",
    "evolve",
]


def _gate_docstring(name, matrix, effect=None) -> str:
    return f"""Apply the {name} gate.
    
    .. csv-table::
        :delim: ;
        :header: Matrix{"; Effect" if effect is not None else ""}

        :math:`{matrix}`{f"; :math:`{effect}`" if effect is not None else ""}
    """


_build_obs = contextvars.ContextVar("build_obs", default=False)


@contextmanager
def obs():
    """
    Context manager to define a observable in Ket.

    When used within a ``with obs():`` block, any operator expressions
    constructed (e.g., sums of Pauli terms) are interpreted as part
    of a observable definition.

    This enables a more natural and symbolic style for building
    observable, closely mirroring their mathematical form.

    Example:

        .. code-block:: python

            with obs():
                h_c = -0.5 * sum(1 - Z(i) * Z(j) for i, j in edges)

    """
    token = _build_obs.set(True)
    try:
        yield
    finally:
        _build_obs.reset(token)


def ham():
    """
    Context manager to define a observable in Ket.

    When used within a ``with obs():`` block, any operator expressions
    constructed (e.g., sums of Pauli terms) are interpreted as part
    of a observable definition.

    This enables a more natural and symbolic style for building
    observable, closely mirroring their mathematical form.

    Example:

        .. code-block:: python

            with obs():
                h_c = -0.5 * sum(1 - Z(i) * Z(j) for i, j in edges)

    """
    warnings.warn(
        "`with ham():` is deprecated and will be removed in future versions."
        " Use `with obs():` instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return obs()


def I(  # pylint: disable=invalid-name missing-function-docstring
    qubits: Quant,
) -> Quant:
    if _build_obs.get():
        return Pauli.i(qubits)

    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    return qubits


I.__doc__ = _gate_docstring(
    "Identity",
    r"\begin{bmatrix} 1 & 0 \\ 0 & 1 \end{bmatrix}",
    r"\begin{matrix} I\left|0\right> = & \left|0\right> \\"
    r"I\left|1\right> = & \left|1\right> \end{matrix}",
)


def X(  # pylint: disable=invalid-name missing-function-docstring
    qubits: Quant,
) -> Quant:
    if _build_obs.get():
        return Pauli.x(qubits)

    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    for qubit in qubits.qubits:
        qubits.process.apply_gate(PAULI_X, 0.0, False, 0, qubit)
    return qubits


X.__doc__ = _gate_docstring(
    "Pauli X",
    r"\begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}",
    r"\begin{matrix} X\left|0\right> = & \left|1\right> \\"
    r"X\left|1\right> = & \left|0\right> \end{matrix}",
)


def Y(  # pylint: disable=invalid-name missing-function-docstring
    qubits: Quant,
) -> Quant:
    if _build_obs.get():
        return Pauli.y(qubits)

    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    for qubit in qubits.qubits:
        qubits.process.apply_gate(PAULI_Y, 0.0, False, 0, qubit)
    return qubits


Y.__doc__ = _gate_docstring(
    "Pauli Y",
    r"\begin{bmatrix} 0 & -i \\ i & 0 \end{bmatrix}",
    r"\begin{matrix} Y\left|0\right> = & i\left|1\right> \\"
    r"Y\left|1\right> = & -i\left|0\right> \end{matrix}",
)


def Z(  # pylint: disable=invalid-name missing-function-docstring
    qubits: Quant,
) -> Quant:
    if _build_obs.get():
        return Pauli.z(qubits)

    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    for qubit in qubits.qubits:
        qubits.process.apply_gate(PAULI_Z, 0.0, False, 0, qubit)
    return qubits


Z.__doc__ = _gate_docstring(
    "Pauli Z",
    r"\begin{bmatrix} 1 & 0 \\ 0 & -1 \end{bmatrix}",
    r"\begin{matrix} Z\left|0\right> = & \left|0\right> \\"
    r"Z\left|1\right> = & -\left|1\right> \end{matrix}",
)


def H(  # pylint: disable=invalid-name missing-function-docstring
    qubits: Quant,
) -> Quant:
    if not isinstance(qubits, Quant):
        qubits = reduce(add, qubits)

    for qubit in qubits.qubits:
        qubits.process.apply_gate(HADAMARD, 0.0, False, 0, qubit)
    return qubits


H.__doc__ = _gate_docstring(
    "Hadamard",
    r"\frac{1}{\sqrt{2}}\begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}",
    r"\begin{matrix}"
    r"H\left|0\right> = & \frac{1}{\sqrt{2}}\left(\left|0\right> + \left|1\right>\right) \\"
    r"H\left|1\right> = & \frac{1}{\sqrt{2}}\left(\left|0\right> - \left|1\right>\right)"
    r"\end{matrix}",
)


def RX(  # pylint: disable=invalid-name missing-function-docstring
    theta: float | Parameter, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:

    def inner(qubits: Quant) -> Quant:
        if not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        for qubit in qubits.qubits:
            if isinstance(theta, Parameter):
                qubits.process.apply_gate(
                    ROTATION_X,
                    theta._multiplier,  # pylint: disable=protected-access
                    True,
                    theta._index,  # pylint: disable=protected-access
                    qubit,
                )
            else:
                qubits.process.apply_gate(
                    ROTATION_X,
                    theta,
                    False,
                    0,
                    qubit,
                )
        return qubits

    if qubits is None:
        return inner
    return inner(qubits)


RX.__doc__ = _gate_docstring(
    "X-axes rotation",
    r"\begin{bmatrix}"
    r"\cos(\theta/2) & -i\sin(\theta/2) \\ -i\sin(\theta/2) & \cos(\theta/2)"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"R_x\left|0\right> = & \cos(\theta/2)\left|0\right> - i\sin(\theta/2)\left|1\right> \\"
    r"R_x\left|1\right> = & -i\sin(\theta/2)\left|0\right> + \cos(\theta/2)\left|1\right>"
    r"\end{matrix}",
)


def RY(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:

    def inner(qubits: Quant) -> Quant:
        if not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        for qubit in qubits.qubits:
            if isinstance(theta, Parameter):
                qubits.process.apply_gate(
                    ROTATION_Y,
                    theta._multiplier,  # pylint: disable=protected-access
                    True,
                    theta._index,  # pylint: disable=protected-access
                    qubit,
                )
            else:
                qubits.process.apply_gate(
                    ROTATION_Y,
                    theta,
                    False,
                    0,
                    qubit,
                )
        return qubits

    if qubits is None:
        return inner
    return inner(qubits)


RY.__doc__ = _gate_docstring(
    "Y-axes rotation",
    r"\begin{bmatrix}"
    r"\cos(\theta/2) & -\sin(\theta/2) \\ \sin(\theta/2) & \cos(\theta/2)"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"R_y\left|0\right> = & \cos(\theta/2)\left|0\right> + \sin(\theta/2)\left|1\right> \\"
    r"R_y\left|1\right> = & -\sin(\theta/2)\left|0\right> + \cos(\theta/2)\left|1\right>"
    r"\end{matrix}",
)


def RZ(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:

    def inner(qubits: Quant) -> Quant:
        if not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        for qubit in qubits.qubits:
            if isinstance(theta, Parameter):
                qubits.process.apply_gate(
                    ROTATION_Z,
                    theta._multiplier,  # pylint: disable=protected-access
                    True,
                    theta._index,  # pylint: disable=protected-access
                    qubit,
                )
            else:
                qubits.process.apply_gate(
                    ROTATION_Z,
                    theta,
                    False,
                    0,
                    qubit,
                )
        return qubits

    if qubits is None:
        return inner
    return inner(qubits)


RZ.__doc__ = _gate_docstring(
    "Z-axes rotation",
    r"\begin{bmatrix} e^{-i\theta/2} & 0 \\ 0 & e^{i\theta/2} \end{bmatrix}",
    r"\begin{matrix} R_z\left|0\right> = & e^{-i\theta/2}\left|0\right> \\"
    r"R_z\left|1\right> = & e^{i\theta/2}\left|1\right> \end{matrix}",
)


def P(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:

    def inner(qubits: Quant) -> Quant:
        if not isinstance(qubits, Quant):
            qubits = reduce(add, qubits)

        for qubit in qubits.qubits:
            if isinstance(theta, Parameter):
                qubits.process.apply_gate(
                    PHASE_SHIFT,
                    theta._multiplier,  # pylint: disable=protected-access
                    True,
                    theta._index,  # pylint: disable=protected-access
                    qubit,
                )
            else:
                qubits.process.apply_gate(
                    PHASE_SHIFT,
                    theta,
                    False,
                    0,
                    qubit,
                )
        return qubits

    if qubits is None:
        return inner
    return inner(qubits)


def PHASE(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:
    warnings.warn(
        "PHASE is deprecated and will be removed in future versions. Use P(theta, qubits) instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return P(theta, qubits)


PHASE.__doc__ = _gate_docstring(
    "Phase shift",
    r"\begin{bmatrix} 1 & 0 \\ 0 & e^{i\theta} \end{bmatrix}",
    r"\begin{matrix} P\left|0\right> = & \left|0\right> \\"
    r"P\left|1\right> = & e^{i\theta}\left|1\right> \end{matrix}",
)


P.__doc__ = _gate_docstring(
    "Phase shift",
    r"\begin{bmatrix} 1 & 0 \\ 0 & e^{i\theta} \end{bmatrix}",
    r"\begin{matrix} P\left|0\right> = & \left|0\right> \\"
    r"P\left|1\right> = & e^{i\theta}\left|1\right> \end{matrix}",
)


S = P(pi / 2)
S.__doc__ = _gate_docstring(
    "S",
    r"\begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}",
    r"\begin{matrix} S\left|0\right> = & \left|0\right> \\"
    r"S\left|1\right> = & i\left|1\right> \end{matrix}",
)

SD = P(-pi / 2)
SD.__doc__ = _gate_docstring(
    "S-dagger",
    r"\begin{bmatrix} 1 & 0 \\ 0 & -i \end{bmatrix}",
    r"\begin{matrix} S^\dagger\left|0\right> = & \left|0\right> \\"
    r"S^\dagger\left|1\right> = & -i\left|1\right> \end{matrix}",
)

T = P(pi / 4)
T.__doc__ = _gate_docstring(
    "T",
    r"\begin{bmatrix} 1 & 0 \\ 0 & e^{i\pi/4} \end{bmatrix}",
    r"\begin{matrix} T\left|0\right> = & \left|0\right> \\"
    r"T\left|1\right> = & e^{i\pi/4}\left|1\right> \end{matrix}",
)

TD = P(-pi / 4)
TD.__doc__ = _gate_docstring(
    "T-dagger",
    r"\begin{bmatrix} 1 & 0 \\ 0 & e^{-i\pi/4} \end{bmatrix}",
    r"\begin{matrix} T^\dagger\left|0\right> = & \left|0\right> \\"
    r"T^\dagger\left|1\right> = & e^{-i\pi/4}\left|1\right> \end{matrix}",
)


def CNOT(  # pylint: disable=invalid-name missing-function-docstring
    control_qubit: Quant, target_qubit: Quant
) -> tuple[Quant, Quant]:
    for c, t in zip(control_qubit, target_qubit):
        ctrl(c, X)(t)
    return control_qubit, target_qubit


CNOT.__doc__ = _gate_docstring(
    "Controlled NOT",
    r"\begin{bmatrix}"
    r"1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \\ 0 & 0 & 1 & 0"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"\text{CNOT}\left|00\right> = & \left|00\right> \\"
    r"\text{CNOT}\left|01\right> = & \left|01\right> \\"
    r"\text{CNOT}\left|10\right> = & \left|11\right> \\"
    r"\text{CNOT}\left|11\right> = & \left|10\right> \\"
    r"\text{CNOT}\left|\text{c}\right>\left|\text{t}\right> ="
    r"& \left|\text{c}\right> \left|\text{c}\oplus\text{t}\right>"
    r"\end{matrix}",
)


def CZ(  # pylint: disable=invalid-name missing-function-docstring
    *qubits: list[Quant],
) -> list[Quant, Quant]:
    for q in zip(*qubits):
        ctrl(q[:-1], Z)(q[-1])
    return qubits


CZ.__doc__ = _gate_docstring(
    "Multi-Controlled Z",
    r"\begin{bmatrix}"
    r"1 & 0 & \cdots & 0 & 0 \\ 0 & 1 & \cdots & 0 & 0 \\"
    r"\vdots & \vdots & \ddots & \vdots & \vdots \\ 0 & 0 & \cdots & 1 & 0 \\"
    r"0 & 0 & \cdots & 0 & -1"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"\text{CZ}\left|0\cdots0\right> = & \left|0\cdots0\right> \\"
    r"\text{CZ}\left|0\cdots1\right> = & \left|0\cdots1\right> \\"
    r"\text{CZ}\left|1\cdots0\right> = & \left|1\cdots0\right> \\"
    r"\text{CZ}\left|1\cdots1\right> = & -\left|1\cdots1\right> \\"
    r"\end{matrix}",
)


def SWAP(  # pylint: disable=invalid-name missing-function-docstring
    qubit_a: Quant, qubit_b: Quant
) -> tuple[Quant, Quant]:
    """Apply a SWAP gate."""
    with around(CNOT, qubit_a, qubit_b):
        CNOT(qubit_b, qubit_a)

    return qubit_a, qubit_b


SWAP.__doc__ = _gate_docstring(
    "SWAP",
    r"\begin{bmatrix}"
    r"1 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"\text{SWAP}\left|00\right> = & \left|00\right> \\"
    r"\text{SWAP}\left|01\right> = & \left|10\right> \\"
    r"\text{SWAP}\left|10\right> = & \left|01\right> \\"
    r"\text{SWAP}\left|11\right> = & \left|11\right> \\"
    r"\text{SWAP}\left|\text{a}\right>\left|\text{b}\right> ="
    r"& \left|\text{b}\right> \left|\text{a}\right>"
    r"\end{matrix}",
)


def RXX(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits_a: Quant | None = None, qubits_b: Quant | None = None
) -> tuple[Quant, Quant] | Callable[[Quant, Quant], tuple[Quant, Quant]]:
    def inner(qubits_a: Quant, qubits_b: Quant) -> tuple[Quant, Quant]:
        for qubit_a, qubit_b in zip(qubits_a, qubits_b):
            with around(cat(kron(H, H), CNOT), qubit_a, qubit_b):
                RZ(theta, qubit_b)

        return qubits_a, qubits_b

    if qubits_a is None and qubits_b is None:
        return inner
    return inner(qubits_a, qubits_b)


RXX.__doc__ = _gate_docstring(
    "XX rotation",
    r"\begin{bmatrix} \cos\frac{\theta}{2} & 0 & 0 & -i\sin\frac{\theta}{2} \\"
    r"0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\"
    r"0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\"
    r"-i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2} \end{bmatrix}",
)


def RZZ(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits_a: Quant | None = None, qubits_b: Quant | None = None
) -> tuple[Quant, Quant] | Callable[[Quant, Quant], tuple[Quant, Quant]]:
    def inner(qubits_a: Quant, qubits_b: Quant) -> tuple[Quant, Quant]:
        for qubit_a, qubit_b in zip(qubits_a, qubits_b):
            with around(CNOT, qubit_a, qubit_b):
                RZ(theta, qubit_b)

        return qubits_a, qubits_b

    if qubits_a is None and qubits_b is None:
        return inner
    return inner(qubits_a, qubits_b)


RZZ.__doc__ = _gate_docstring(
    "ZZ rotation",
    r"\begin{bmatrix} e^{-i \frac{\theta}{2}} & 0 & 0 & 0 \\"
    r"0 & e^{i \frac{\theta}{2}} & 0 & 0\\"
    r" 0 & 0 & e^{i \frac{\theta}{2}} & 0 \\"
    r"0 & 0 & 0 & e^{-i \frac{\theta}{2}} \end{bmatrix}",
)


def RYY(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits_a: Quant | None = None, qubits_b: Quant | None = None
) -> tuple[Quant, Quant] | Callable[[Quant, Quant], tuple[Quant, Quant]]:
    def inner(qubits_a: Quant, qubits_b: Quant) -> tuple[Quant, Quant]:
        for qubit_a, qubit_b in zip(qubits_a, qubits_b):
            with around(cat(kron(RX(pi / 2), RX(pi / 2)), CNOT), qubit_a, qubit_b):
                RZ(theta, qubit_b)
        return qubits_a, qubits_b

    if qubits_a is None and qubits_b is None:
        return inner
    return inner(qubits_a, qubits_b)


RYY.__doc__ = _gate_docstring(
    "RYY rotation",
    r"\begin{bmatrix}"
    r"\cos\frac{\theta}{2} & 0 & 0 & i\sin\frac{\theta}{2} \\"
    r"0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\"
    r"0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\"
    r"i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2}"
    r"\end{bmatrix}",
)


def RBS(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, qubits_a: Quant | None = None, qubits_b: Quant | None = None
) -> tuple[Quant, Quant] | Callable[[Quant, Quant], tuple[Quant, Quant]]:
    def inner(qubits_a: Quant, qubits_b: Quant) -> tuple[Quant, Quant]:
        for qubit_a, qubit_b in zip(qubits_a, qubits_b):
            with around(cat(kron(H, H), CNOT), qubit_a, qubit_b):
                RY(theta / 2, qubit_a)
                RY(-theta / 2, qubit_b)
        return qubits_a, qubits_b

    if qubits_a is None and qubits_b is None:
        return inner
    return inner(qubits_a, qubits_b)


RBS.__doc__ = _gate_docstring(
    "Reconfigurable Beam Splitter (RBS)",
    r"\begin{bmatrix}"
    r"1 & 0 & 0 & 0 \\"
    r"0 & \cos\theta & \sin\theta & 0 \\"
    r"0 & -\sin\theta & \cos\theta & 0 \\"
    r"0 & 0 & 0 & 1"
    r"\end{bmatrix}",
)


def U3(  # pylint: disable=invalid-name missing-function-docstring
    theta: float, phi: float, lambda_: float, qubit: Quant | None = None
) -> Quant | Callable[[Quant], Quant]:
    gate = cat(RZ(lambda_), RY(theta), RZ(phi))
    if qubit is not None:
        return gate(qubit)
    return gate


U3.__doc__ = _gate_docstring(
    "U3",
    r"\begin{bmatrix}"
    r"e^{-i (\phi + \lambda)/2} \cos(\theta/2) & -e^{-i (\phi - \lambda)/2} \sin(\theta/2) \\"
    r"e^{i (\phi - \lambda)/2} \sin(\theta/2) & e^{i (\phi + \lambda)/2} \cos(\theta/2)"
    r"\end{bmatrix}",
    r"\begin{matrix}"
    r"U3\left|0\right> = & e^{-i (\phi + \lambda)/2} \cos(\theta/2)\left|0\right>"
    r"+ e^{i (\phi - \lambda)/2} \sin(\theta/2) \left|1\right> \\"
    r"U3\left|1\right> = & -e^{-i (\phi - \lambda)/2} \sin(\theta/2)\left|0\right>"
    r"+ e^{i (\phi + \lambda)/2} \cos(\theta/2)\left|1\right> \\"
    r"\end{matrix}",
)


def global_phase(
    theta: float,
) -> Callable[[Callable[[Any], Any]], Callable[[Any], Any]]:
    r"""Apply a global phase to a quantum operation.

    Decorator that adds a global phase :math:`e^{i\theta}` to a quantum gate
    :math:`U`, creating the gate :math:`e^{i\theta}U`.

    In quantum computation, global phases are overall factors that can be
    applied to quantum states without affecting the measurement outcomes.
    Mathematically, they represent rotations in the complex plane and are
    usually ignored because they have no observable consequences. However, in
    certain contexts, such as controlled quantum operations, the global phase
    can affect the behavior of the operation.

    The addition of a global phase can be important for maintaining
    consistency in quantum algorithms, particularly when dealing with controlled
    operations where relative phase differences between different components
    of the quantum state can impact the computation.

    Example:

    .. code-block:: python

        @global_phase(pi / 2)
        def my_z_gate(qubit):
            return RZ(pi, qubit)

    This example defines a custom quantum gate equivalent to a Pauli Z
    operation, where :math:`Z = e^{i\frac{\pi}{2}}R_z(\pi)`.

    Args:
        theta: The :math:`\theta` angle of the global phase :math:`e^{i\theta}`.
    """

    def _global_phase(gate: Callable[[Any], Any]) -> Callable[[Any], Any]:
        @functools.wraps(gate)
        def inner(*args, ket_process: Process | None = None, **kwargs):
            ket_process = _search_process(ket_process, args, kwargs)

            ket_process.apply_global_phase(theta)

            return gate(*args, **kwargs)

        return inner

    return _global_phase


SX = global_phase(pi / 4)(RX(pi / 2))  # pylint: disable=invalid-name
SX.__doc__ = _gate_docstring(
    "Sqrt X",
    r"\frac{1}{2} \begin{bmatrix} 1+i & 1-i \\ 1-i & 1+i \end{bmatrix}",
    r"\begin{matrix}"
    r"\sqrt{X}\left|0\right> = & \frac{1}{2} ((1+i)\left|0\right> + (1-i)\left|1\right>) \\"
    r"\sqrt{X}\left|1\right> = & \frac{1}{2} ((1-i)\left|0\right> + (1+i)\left|1\right>)"
    r"\end{matrix}",
)


def QFT(qubits, do_swap: bool = True):  # pylint: disable=invalid-name
    r"""Quantum Fourier Transform.

    .. math::

        \text{QFT}\left|x\right> =
        \frac{1}{\sqrt{N}} \sum_{k=0}^{N-1} e^{2\pi i x k / N} \left|k\right>

    args:
        qubits: The qubits to apply the QFT to.
        do_swap: Whether to invert the qubits after the transformation.
    """
    if len(qubits) == 1:
        H(qubits)
    else:
        *init, last = qubits
        H(last)

        for i, ctrl_qubit in enumerate(reversed(init)):
            with control(ctrl_qubit):
                P(pi / 2 ** (i + 1), last)

        QFT(init, do_swap=False)

    if do_swap:
        size = len(qubits)
        for i in range(size // 2):
            SWAP(qubits[i], qubits[size - i - 1])


_EVOLVE_GATES = {
    "X": (RX, RXX),
    "Y": (RY, RYY),
    "Z": (RZ, RZZ),
    "I": (lambda *_: ..., lambda *_: ...),
}


def evolve(hamiltonian: Hamiltonian):
    """Evolve the quantum state according to the given Hamiltonian."""
    if max(map(len, hamiltonian.pauli_products)) > 2:
        raise ValueError("Only 1- and 2-qubit terms are supported in evolve.")
    for pauli in hamiltonian.pauli_products:
        if len(pauli) <= 1:
            continue
        gates = list(pauli.map.values())
        if gates[0] != gates[1]:
            raise ValueError("Only XX, YY, and ZZ interactions are supported")

    process = hamiltonian.process
    for pauli in hamiltonian.pauli_products:
        if len(pauli) == 0:
            continue
        qubits = Quant(qubits=list(pauli.map.keys()), process=process)
        gates = list(pauli.map.values())[0]
        _EVOLVE_GATES[gates][len(pauli) - 1](2 * pauli.coef, *qubits)



"""Quantum library.

Utilities for preparing quantum states and building quantum algorithms.
"""

from __future__ import annotations

# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

from functools import reduce
from operator import add
from typing import Callable, Literal
from math import sqrt
from collections.abc import Iterable
from inspect import signature

from ..clib.libket import BatchExecution
from ..base import Process
from ..operations import dump
from ..gates import H, CNOT
from . import gates, prepare, math, oracle, ham

try:
    import google.colab  # pylint: disable=unused-import

    _in_notebook = True
except ImportError:
    try:
        from IPython import get_ipython

        _in_notebook = get_ipython().__class__.__name__ == "ZMQInteractiveShell"

    except ImportError:
        _in_notebook = False

try:
    from qiskit import QuantumCircuit, QuantumRegister
    from qiskit.circuit import library

    QISKIT_AVAILABLE = True
except ImportError:
    QISKIT_AVAILABLE = False

__all__ = [
    "draw",
    "dump_matrix",
    "gates",
    "ham",
    "math",
    "oracle",
    "prepare",
]


def dump_matrix(
    gate: Callable,
    num_qubits: int | list[int] = 1,
    args=(),
    process: Process | None = None,
) -> list[list[complex]]:
    """Get the matrix representation of a quantum gate.

    This function calculates the matrix representation of a quantum gate.

    Args:
        gate: Quantum gate operation to obtain the matrix for.
        num_qubits: Number of qubits.
        args: Classical arguments to pass to the gate function.
        process: Quantum process used to generate the matrix.

    Returns:
        Matrix representation of the quantum gate.
    """

    if isinstance(num_qubits, Iterable):
        qubit_args = num_qubits
    else:
        qubit_args = [num_qubits]

    num_qubits = sum(num_qubits) if isinstance(num_qubits, Iterable) else num_qubits

    if process is None:
        process = Process(
            num_qubits=2 * num_qubits,
            execution="batch",
            simulator="sparse",
        )

    mat = [[0.0j for _ in range(2**num_qubits)] for _ in range(2**num_qubits)]

    qubit_args = [process.alloc(n) for n in qubit_args]
    row = reduce(add, qubit_args)
    column = process.alloc(num_qubits)

    H(column)
    CNOT(column, row)

    gate(*args, *qubit_args)

    state = dump(column + row)

    for state, amp in state.get().items():
        column = state >> num_qubits
        row = state & ((1 << num_qubits) - 1)
        mat[row][column] = amp * sqrt(2**num_qubits)

    return mat


DRAW_STYLE = {
    "textcolor": "#0f1f2e",
    "gatetextcolor": "#f4f8ff",
    "subtextcolor": "#f4f8ff",
    "linecolor": "#0f1f2e",
    "creglinecolor": "#6b7682",
    "gatefacecolor": "#0f7cfb",
    "barrierfacecolor": "#c6cdd5",
    "backgroundcolor": "#f4f8ff",
    "margin": [1.5, 0, 0, 0],
    "displaytext": {
        "u1": "P",
        "rx": "RX",
        "ry": "RY",
        "rz": "RZ",
    },
    "displaycolor": {
        "u1": ["#008699", "#f4f8ff"],
        "x": ["#0f7cfb", "#f4f8ff"],
        "y": ["#0c63c9", "#f4f8ff"],
        "z": ["#009cb3", "#f4f8ff"],
        "h": ["#3f96fc", "#f4f8ff"],
        "rx": ["#063264", "#f4f8ff"],
        "ry": ["#094a97", "#f4f8ff"],
        "rz": ["#008699", "#f4f8ff"],
        "cx": ["#0f7cfb", "#f4f8ff"],
        "ccx": ["#0f7cfb", "#f4f8ff"],
        "mcx": ["#0f7cfb", "#f4f8ff"],
        "cy": ["#0c63c9", "#f4f8ff"],
        "cz": ["#009cb3", "#f4f8ff"],
    },
}


class _IBMDeviceForDraw(BatchExecution):
    """IBMDevice use for qulib.draw"""

    def __init__(
        self,
        num_qubits: list[int],
        names: list[str],
        qpu: dict | None = None,
        keep_order: bool = True,
    ):
        super().__init__()
        qubits = [QuantumRegister(n, l) for n, l in zip(num_qubits, names)]
        self.circuit = QuantumCircuit(*qubits)
        self.num_qubits = sum(num_qubits)
        self.qpu = qpu
        self.keep_order = keep_order
        self.last_gate = None

    def clear(self):
        pass

    def submit_execution(self, circuit, _p):
        self.process_instructions(circuit)

    def get_result(self):
        return {
            "measurements": [],
            "exp_values": [],
            "samples": [],
            "dumps": [],
            "gradients": None,
        }

    def pauli_x(self, target, control):
        if self.keep_order and self.last_gate != "X":
            self.circuit.barrier()
        gate = library.XGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "X"

    def pauli_y(self, target, control):
        if self.keep_order and self.last_gate != "Y":
            self.circuit.barrier()
        gate = library.YGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "Y"

    def pauli_z(self, target, control):
        if self.keep_order and self.last_gate != "Z":
            self.circuit.barrier()
        gate = library.ZGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "Z"

    def hadamard(self, target, control):
        if self.keep_order and self.last_gate != "H":
            self.circuit.barrier()
        gate = library.HGate()
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "H"

    def rotation_x(self, target, control, **kwargs):
        if self.keep_order and self.last_gate != "RX":
            self.circuit.barrier()
        gate = library.RXGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "RX"

    def rotation_y(self, target, control, **kwargs):
        if self.keep_order and self.last_gate != "RY":
            self.circuit.barrier()
        gate = library.RYGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "RY"

    def rotation_z(self, target, control, **kwargs):
        if self.keep_order and self.last_gate != "RZ":
            self.circuit.barrier()
        gate = library.RZGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "RZ"

    def phase(self, target, control, **kwargs):
        if self.keep_order and self.last_gate != "P":
            self.circuit.barrier()
        gate = library.PhaseGate(kwargs["Value"])
        if len(control) >= 1:
            gate = gate.control(len(control))
        self.circuit.append(gate, control + [target])
        self.last_gate = "P"

    def connect(
        self,
    ):
        """Configure process"""
        return super().configure(
            num_qubits=self.num_qubits,
            execution_managed_by_target={},
            qpu=self.qpu,
        )


def draw(  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
    gate: Callable,
    qubits: int | list[int],
    args: tuple = (),
    *,
    qpu_size: int | None = None,
    u4_gate: Literal["CX", "CZ"] | None = None,
    u2_gates: Literal["ZYZ", "RzSx"] | None = None,
    coupling_graph: list[tuple[int, int]] | None = None,
    title: str | None = None,
    keep_order: bool = True,
    **kwargs,
):
    """Draw a quantum gate using Qiskit.

    Note:
        This method requires additional dependencies from ``ket-lang[plot]``.

        Install with: ``pip install ket-lang[plot]``.

    Args:
        gate: Quantum gate function.
        qubits: Number of qubits.
        args: Classical arguments to pass to the gate function.
        qpu_size: Size of the quantum processing unit (QPU).
            If specified, the number of qubits will be adjusted to fit the QPU size.
        u4_gate: Type of U4 gate to use, either "CX" or "CZ".
        u2_gates: Type of U2 gates to use, either "ZYZ" or "RzSx".
        coupling_graph: Coupling graph of the QPU,
            specified as a list of tuples representing connected qubits.
        title: Title for the circuit diagram.
        keep_order: Maintain the gate call order.
        **kwargs: Keyword arguments to pass to the Qiskit drawer.

    Returns:
        Qiskit circuit diagram of the quantum gate.
    """

    if not QISKIT_AVAILABLE:
        raise RuntimeError(
            "Visualization optional dependence are required. Install with: "
            "pip install ket-lang[plot]"
        )

    if not isinstance(qubits, Iterable):
        qubits = [qubits]
    if not isinstance(args, Iterable):
        args = [args]

    names = list(signature(gate).parameters)[len(args) : len(args) + len(qubits)]
    if len(names) != len(qubits):
        names = [None for _ in range(len(qubits))]

    if qpu_size is not None:
        qubits = list(qubits)
        total = sum(qubits)
        if total > qpu_size:
            raise ValueError(
                f"Total number of qubits {sum(qubits)} exceeds the QPU size {qpu_size}"
            )
        qubits.append(qpu_size - total)
        names.append("AUX")

    qpu = {}

    if u4_gate is not None:
        if not u4_gate in ["CX", "CZ"]:
            raise ValueError("u4_gate must be 'CX' or 'CZ'")
        qpu["u4_gate"] = u4_gate
    if u2_gates is not None:
        if not u2_gates in ["ZYZ", "RzSx"]:
            raise ValueError("u2_gates must be 'ZYZ' or 'RzSx'")
        qpu["u2_gates"] = u2_gates
    if coupling_graph is not None:
        qpu["coupling_graph"] = coupling_graph

    device = _IBMDeviceForDraw(
        [qpu_size] if coupling_graph is not None else qubits,
        ["Q"] if coupling_graph is not None else names,
        qpu if len(qpu) else None,
        keep_order,
    )

    p = Process(device)
    q = [p.alloc(n) for n in (qubits if qpu_size is None else qubits[:-1])]
    gate(*args, *q)
    p.execute()

    if "output" not in kwargs and _in_notebook:
        kwargs["output"] = "mpl"

    kwargs["style"] = {**DRAW_STYLE, **kwargs.get("style", {})}

    fig = device.circuit.draw(**kwargs, plot_barriers=False)
    if title is not None:
        fig.suptitle(title)
    return fig



# SPDX-FileCopyrightText: 2020 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
# SPDX-FileCopyrightText: 2020 Rafael de Santiago <r.santiago@ufsc.br>
#
# SPDX-License-Identifier: Apache-2.0

"""Phase estimation algorithm"""

from math import pi
from functools import partial
import ket


def qft(qubits: ket.Quant, invert: bool = True):
    """
    Applies the Quantum Fourier Transform (QFT) to the given qubits.

    Args:
        qubits (quant): A quantum register containing the qubits to be transformed.
        invert (bool): Whether inverse the qubits order at the end of the QFT. Default is True.
    """
    if len(qubits) == 1:
        # Base case:
        # single-qubit QFT is simply the Hadamard gate.
        ket.H(qubits)
    else:
        # Recursive case:
        # Apply a Hadamard gate to the last qubit.
        *head, tail = qubits
        ket.H(tail)

        # Apply controlled phase shift gates to the last qubit with each of the remaining qubits.
        for i, ctrl_qubit in enumerate(reversed(head)):
            with ket.control(ctrl_qubit):
                ket.P(pi / 2 ** (i + 1), tail)

        # Recursively apply QFT to the remaining qubits.
        qft(head, invert=False)

    # Swap the order of the qubits.
    if invert:
        size = len(qubits)
        for i in range(size // 2):
            ket.SWAP(qubits[i], qubits[size - i - 1])


def phase_estimator(oracle_gate, precision: int) -> float:
    """Estimates the phase of a given oracle function to a certain precision
    using the phase estimation algorithm.

    Args:
        oracle (callable): A function that performs a phase shift operation on
            the target qubit based on a given phase and index value.
            It should take the form: `oracle(phase_: float, i: int, tgr: quant) -> None`
        precision (int): The number of bits of precision desired in the estimation.

    Returns:
        float: An float value representing the estimated phase.
    """

    p = ket.Process(simulator="dense", num_qubits=precision + 1)

    # Create a quantum register to hold the control qubits.
    ctr = ket.H(p.alloc(precision))

    # Create a target qubit.
    tgr = ket.X(p.alloc())

    # Apply the oracle function with each control qubit.
    for i, c in enumerate(ctr):
        with ket.control(c):
            oracle_gate(i, tgr)

    # Apply the inverse QFT to the control qubits.
    ket.adj(qft)(ctr)

    # Measure the control qubits and return the result.
    return ket.measure(reversed(ctr)).value / 2**precision


def oracle(phase: float, i: int, tgr):
    """Performs a phase shift operation on the target
    qubit based on a given phase and index value.

    Args:
        phase_ (float): The phase value to be applied.
        i (int): The index value.
        tgr (quant): The target qubit.
    """
    ket.P(2 * pi * phase * 2**i, tgr)


estimate_pi = partial(phase_estimator, partial(oracle, pi / 10))


def test_phase_estimator():
    assert abs((estimate_pi(16) * 10) - pi) < 1e-3


if __name__ == "__main__":
    test_phase_estimator()
    print("OK")



"""QuForge Interface for Ket

The class :class:`~ket.quforge.QuForgeKet` enables the use of QuForge, an efficient
qudit simulator, within Ket. This simulator supports expected value calculation
and gradient evaluation, making it a great choice for Quantum Machine Learning.

This module requires additional dependencies from ``ket-lang[qml]``.
Install with: ``pip install ket-lang[qml]``.

"""

# SPDX-FileCopyrightText: 2025 Evandro Chagas Ribeiro da Rosa <evandro@quantuloop.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Literal
from functools import reduce, partial
from .clib.libket import BatchExecution

try:
    from quforge import quforge as qf
    import torch

    QUFORGE_AVAILABLE = True
except ImportError:
    QUFORGE_AVAILABLE = False


class QuForgeKet(BatchExecution):  # pylint: disable=too-many-instance-attributes
    """QuForge Interface for Ket

    This simulator supports expected value calculation and gradient evaluation,
    making it a great choice for Quantum Machine Learning.

    Args:
         num_qubits: Number of qubits to simulate.
         device: Specifies whether to run the simulation on the CPU or GPU. Defaults to "cpu".
         sparse: If True, uses a sparse representation for improved performance on large systems.
             Defaults to True.
         gradient: Enables gradient evaluation for optimization tasks. Defaults to True.
    """

    def __init__(
        self,
        num_qubits: int,
        device: Literal["cpu", "cuda"] = "cpu",
        sparse: bool = True,
        gradient: bool = True,
    ):
        if not QUFORGE_AVAILABLE:
            raise RuntimeError(
                "QuForge is not available. Please install it with: pip install ket-lang[qml]"
            )

        super().__init__()

        self.initial_state = qf.State(
            "-".join(["0"] * num_qubits), dim=2, device=device
        )

        self.num_qubits = num_qubits
        self.device = device
        self.sparse = sparse
        self.gradient_enabled = gradient

        self.p_z = qf.Z(dim=2, device=device)
        self.p_x = qf.X(dim=2, device=device)
        self.p_y = qf.Y(dim=2, device=device)
        self.p_i = qf.eye(dim=2, device=device, sparse=sparse)

        if sparse:
            self.p_z = self.p_z.M.to_sparse()
            self.p_x = self.p_x.M.to_sparse()
            self.p_y = self.p_y.M.to_sparse()

        self.pauli_map = {
            "X": self.p_x,
            "Y": self.p_y,
            "Z": self.p_z,
            "I": self.p_i,
        }

        self.circuit = None
        self.exp_result = None
        self.gradient = None
        self.parameters = None

    def pauli_x(self, target, control):
        assert len(control) <= 1, "Control qubits are not supported"
        if len(control) == 0:
            self.circuit.X(index=[target])
        else:
            self.circuit.CNOT(index=[control[0], target])

    def pauli_y(self, target, control):
        assert len(control) == 0, "Control qubits are not supported"
        self.circuit.Y(index=[target])

    def pauli_z(self, target, control):
        assert len(control) == 0, "Control qubits are not supported"
        self.circuit.Z(index=[target])

    def hadamard(self, target, control):
        assert len(control) == 0, "Control qubits are not supported"
        self.circuit.H(index=[target])

    def rotation_x(self, target, control, **kwargs):
        assert len(control) == 0, "Control qubits are not supported"
        match kwargs:
            case {"Value": value}:
                ...
            case {"Ref": {"index": index, "multiplier": multiplier, "value": _}}:
                value = self.parameters[index] * multiplier

        self.circuit.RX(index=[target], angle=value)

    def rotation_y(self, target, control, **kwargs):
        assert len(control) == 0, "Control qubits are not supported"
        match kwargs:
            case {"Value": value}:
                ...
            case {"Ref": {"index": index, "multiplier": multiplier, "value": _}}:
                value = self.parameters[index] * multiplier

        self.circuit.RY(index=[target], angle=value)

    def rotation_z(self, target, control, **kwargs):
        assert len(control) == 0, "Control qubits are not supported"
        match kwargs:
            case {"Value": value}:
                ...
            case {"Ref": {"index": index, "multiplier": multiplier, "value": _}}:
                value = self.parameters[index] * multiplier

        self.circuit.RZ(index=[target], angle=value)

    def phase(self, target, control, **kwargs):
        self.rotation_z(target, control, **kwargs)

    def exp_value(self, _, hamiltonian):
        coef = hamiltonian["coefficients"]
        products = hamiltonian["products"]

        h_final = None

        for pauli, c in zip(products, coef):
            h_partial = ["I" for _ in range(self.num_qubits)]
            for p in pauli:
                h_partial[p["qubit"]] = p["pauli"][-1]
            h_partial = list(map(lambda x: self.pauli_map[x], h_partial))
            h_partial = c * reduce(partial(qf.kron, sparse=self.sparse), h_partial)

            if h_final is None:
                h_final = h_partial
            else:
                h_final += h_partial

        output_state = self.circuit(self.initial_state)
        expected_value = (output_state.T.conj() @ (h_final @ output_state)).real
        if self.gradient_enabled and len(self.parameters):
            expected_value.backward()
        self.exp_result = float(expected_value[0])

        self.gradient = [p.grad.sum().item() for p in self.parameters]

    def submit_execution(self, circuit, parameters):
        self.gradient = [None for _ in range(len(parameters))]
        self.parameters = [
            torch.nn.Parameter(p * torch.ones(self.num_qubits, device=self.device))
            for p in parameters
        ]
        self.process_instructions(circuit)

    def get_result(self):
        return {
            "measurements": [],
            "exp_values": [self.exp_result] if self.exp_result is not None else [],
            "samples": [],
            "dumps": [],
            "gradients": self.gradient if self.gradient_enabled else None,
        }

    def clear(self):
        self.circuit = qf.Circuit(
            dim=2,
            wires=self.num_qubits,
            sparse=self.sparse,
            device=self.device,
        )

        self.exp_result = None
        self.gradient = None
        self.parameters = None

    def connect(self):
        self.clear()
        return super().configure(
            num_qubits=self.num_qubits,
            execution_managed_by_target={"exp_value": "Basic"},
            qpu={"u2_gates": "ZYZ"},
            gradient="NativeSupport" if self.gradient_enabled else None,
        )
